import torch

from torch.distributions import Normal, Uniform

from infer import run_inference, importance_resample
from ppl import ProbCtx, run_prob_prog

distance_limit = 10


def walk_model(ctx: ProbCtx) -> float:
    """Random walk model.

    Mak et al. (2020): Densities of almost-surely terminating probabilistic programs are differentiable almost everywhere.
    """
    distance = torch.tensor(0.0, requires_grad=True)
    start = ctx.sample(Uniform(0, 3), is_cont=False)
    position = start
    while position > 0 and distance < distance_limit:
        step = ctx.sample(Uniform(-1, 1), is_cont=False)
        distance = distance + torch.abs(step)
        position = position + step
    ctx.observe(distance, Normal(1.1, 0.1))
    return start.item()


if __name__ == "__main__":
    count = 1_000
    repetitions = 10
    configs = [
        (L, alpha, K, eps)
        for L in [5, 10, 25]
        for eps in [0.2, 0.1]
        for alpha in [0.5, 0.1]
        for K in [0, 1, 2]
    ]
    for rep in range(repetitions):
        for L, alpha, K, eps in configs:
            print(f"REPETITION {rep+1}/{repetitions} ({eps=}, {L=}, {alpha=}, {K=})")
            run_inference(
                lambda trace: run_prob_prog(walk_model, trace=trace),
                name=f"walk_model_{rep}",
                count=count,
                burnin=0,  # 100,
                eps=eps,
                L=L,
                K=K,
                alpha=alpha,
                seed=rep,
            )
